import Numeric
from wxPython.wx import *
from wxPython.glcanvas import *
from OpenGL.GL import *
from OpenGL.GLUT import *
from OpenGL.GLU import *
import sys
import copy
import threading
import math
from sets import Set
import time
import profile

from GLWindow import GLWindow
from Module import Module
from ResourceManager import ResourceManager
from Progressor import Progressor

import globals

SINGLESELECTIONEVENT = wxNewEventType()

def EVT_MAPSINGLESELECTION(window,function):
    '''notifies about the selection of a single object in the map'''
    window.Connect(-1,-1,SINGLESELECTIONEVENT,function)

class SingleSelectionEvent(wxPyCommandEvent):
    eventType = SINGLESELECTIONEVENT
    def __init__(self,windowID,oid):
        wxPyCommandEvent.__init__(self,self.eventType,windowID)
        self.objectID = oid

    def getSelectedId(self):
        return self.objectID
    
    def Clone(self):
        self.__class__(self.GetId(),self.objectID)

MOVEEVENT = wxNewEventType()
def EVT_MAPMOVE(window,function):
    '''notifies about movement of objects in the map'''
    window.Connect(-1,-1,MOVEEVENT,function)

class MoveEvent(wxPyCommandEvent):
    eventType = MOVEEVENT
    def __init__(self,windowID,oid,x,y):
        wxPyCommandEvent.__init__(self,self.eventType,windowID)
        self.objectID = oid
        self.x = x
        self.y = y

    def getObjectId(self):
        return self.oid

    def getX(self):
        return self.x

    def getY(self):
        return self.y

    def Clone(self):
        self.__class__(self.GetId(),self.objectID,self.x,self.y)

    
class MapWindow(GLWindow,Progressor):
    def __init__(self,parent):
        GLWindow.__init__(self, parent)
        Progressor.__init__(self)
        
        self.zoom = 20

        self.players = {}
        self.area = None
        self.placeables = None
        self.doors = None
        self.creatures = None
        self.lock = threading.Lock()
        self.highlight = None
        self.highlightBox = None
        self.beingDragged = None
        self.selected = []

        self.fps = 0.0
        self.showFPS = False
        self.preprocessed = False
        
    def OnMouseDown(self, evt):
        if self.highlight != None:
            self.selected = [self.highlight]
            self.beingDragged = self.getThingHit(self.highlight)
            event = SingleSelectionEvent(self.GetId(),self.beingDragged.getNevereditId())
            self.GetEventHandler().AddPendingEvent(event)
            self.dragStart = (evt.GetX(),self.height-evt.GetY())
            self.setupCamera()
            x,y,z = self.project(self.beingDragged.getX(),
                                 self.beingDragged.getY(),
                                 0.0)
            self.dragOffset = (self.dragStart[0] - x,
                               self.dragStart[1] - y)
        else:
            self.selected = []
        self.DrawGLScene() #Refresh(False)

    def OnMouseUp(self, evt):
        self.beingDragged = None

    def OnKeyDown(self,evt):
        GLWindow.OnKeyDown(self,evt)        
        if evt.GetKeyCode() == WXK_SPACE:
            print 'profiling to draw.prof'
            import profilewrap
            f = profilewrap.wrap(self.DrawGLScene)
            p = profile.Profile()
            p.runcall(f,self)
            p.dump_stats('draw.prof')

    def getThingHit(self,name):
        if name < len(self.doors):
            return self.doors[name]
        name -= len(self.doors)
        if name < len(self.placeables):
            return self.placeables[name]
        name -= len(self.placeables)
        if name < len(self.creatures):
            return self.creatures[name]
        return None

    def selectThingById(self,id):
        didSelect = False
        for i,thing in enumerate(self.doors + self.placeables + self.creatures):
            if thing.getNevereditId() == id:
                if i not in self.selected:
                    self.highlight = None
                    self.selected = [i]
                    self.centreThing(thing)
                    self.Refresh(False)
                didSelect = True
        if not didSelect:
            print >>sys.stderr,__name__,'cannot find thing with id',id
        

    def centreThing(self,thing):
        self.lookingAtX = thing.getX()
        self.lookingAtY = thing.getY()
        
    def OnMouseMotion(self, evt):
        if not self.area or not self.preprocessed:
            return
        if not self.beingDragged:
            buf = self.doSelection(evt.GetX(),self.height-evt.GetY())
            for hit_record in buf:
                min_depth,max_depth,names = hit_record
                if len(names):
                    self.setHighlightBox(names[0])
                    break
            else:
                self.highlight = None
            self.SetupProjection()
            self.DrawGLScene()#Refresh(False)
        else:
            glPushMatrix()
            self.setupCamera()
            dragX = float(evt.GetX() - self.dragOffset[0])
            dragY = float(self.height-evt.GetY() - self.dragOffset[1])
            near = Numeric.array(self.unproject(dragX,dragY,0.0))
            far = Numeric.array(self.unproject(dragX,dragY,1.0))
            N = Numeric.array([0,0,1]) #this is normal to the area plane
            P = Numeric.array([1,1,0.01]) #this is on the area plane
            u = Numeric.dot(N,P-near)/Numeric.dot(N,far-near)
            intersect = near + u * (far-near)
            x = intersect[0]
            y = intersect[1]
            glPopMatrix()
            if x < 0:
                x = 0
            if x > self.area.getWidth()*10.0:
                x = self.area.getWidth()*10.0
            if y < 0:
                y = 0
            if y > self.area.getHeight()*10.0:
                y = self.area.getHeight()*10.0                
            self.beingDragged.setX(x)
            self.beingDragged.setY(y)
            event = MoveEvent(self.GetId(),self.beingDragged.getNevereditId(),x,y)
            self.GetEventHandler().AddPendingEvent(event)

            self.DrawGLScene() #self.Refresh(False)

    def clearCache(self):
        GLWindow.clearCache(self)
        self.preprocessedModels = Set()
        
    def preprocess(self):        
        self.clearCache()
        self.setStatus("Preparing door display...")
        self.setProgress(10)
        for i,d in enumerate(self.doors):
            if d.getModel():
                if not d.modelName in self.preprocessedModels:
                    self.preprocessNodes(d.getModel(),'d'+`i`,bbox=True)
                    self.preprocessedModels.add(d.modelName)
            else:
                print 'warning, no model for',d.getName()
        self.setProgress(30)
        self.setStatus("Preparing placeable display...")
        for i,p in enumerate(self.placeables):
            if p.getModel():
                if not p.modelName in self.preprocessedModels:
                    self.preprocessNodes(p.getModel(),'p'+`i`,bbox=True)
                    self.preprocessedModels.add(p.modelName)
            else:
                print 'warning, no model for',p.getName()
        self.setProgress(50)
        self.setStatus("Preparing creature display...")
        for i,c in enumerate(self.creatures):
            if c.getModel():
                if not c.modelName in self.preprocessedModels:
                    self.preprocessNodes(c.getModel(),'c'+`i`,bbox=True)
                    self.preprocessedModels.add(c.modelName)                
            else:
                pass # I know I'm not handling all model types here yet
        self.setProgress(70)
        self.setStatus("Preparing tile display...")
        for i,t in enumerate(self.tiles):
            if t.getModel():
                if not t.modelName in self.preprocessedModels:
                    self.preprocessNodes(t.getModel(),'t'+`i`,bbox=True)
                    self.preprocessedModels.add(t.modelName)                
            else:
                print 'warning, no model for',t.getName()
        self.setProgress(0)
        self.setStatus("Map display prepared.")

    def setArea(self,area):
        if self.area == area:
            return
        self.area = None
        self.lock.acquire()
        self.selected = []
        self.highlight = None
        if not area:
            self.selected = []
            self.placeables = None
            self.doors = None
            self.creatures = None
            self.tiles = None
        else:
            self.setProgress(0)
            self.setStatus('Reading area contents...')
            area.readContents()
            self.setProgress(50)
            self.setStatus('Reading area tiles...')
            area.readTiles()
            self.lookingAtX = area.getWidth()*5.0
            self.lookingAtY = area.getWidth()*5.0        
            self.placeables = area.getPlaceables()
            self.doors = area.getDoors()
            self.creatures = area.getCreatures()
            self.tiles = area.getTiles()
            self.preprocessed = False
            self.setProgress(0)
        self.area = area
        self.lock.release()
        self.Refresh(True)

    def getBaseWidth(self):
        if self.area:
            return self.area.getWidth()*10.0
        else:
            return 1.0

    def getBaseHeight(self):
        if self.area:
            return self.area.getHeight()*10.0
        else:
            return 1.0

    def setPlayer(self,p):
        self.lock.acquire()
        self.players[p.playerID2] = copy.copy(p)
        self.lock.release()
        self.Refresh(False)

    def removePlayer(self,pid):
        self.lock.acquire()
        print 'removing player',pid
        del self.players[pid]
        self.lock.release()
        self.Refresh(False)

    def setHighlightBox(self,hid):
        self.highlight = hid
        class HighlightBox:
            pass
        self.highlightBox = HighlightBox()
        target = self.getThingHit(hid)
        if target:
            self.highlightBox.text = target.getName()# + ' ' + `int(target.getObjectId())`
        else:
            self.highlightBox.text = ''            
        self.highlightBox.red = 0.3
        self.highlightBox.blue = 0.3
        self.highlightBox.green = 0.3
        self.highlightBox.alpha = 0.35
        
        
    def drawOverlays(self):
        glDisable(GL_DEPTH_TEST)
        glDisable(GL_LIGHTING)
        glMatrixMode(GL_PROJECTION)
        glPushMatrix()
        glLoadIdentity()
        gluOrtho2D(0,self.width,0,self.height)
        glMatrixMode(GL_MODELVIEW)
        glPushMatrix()
        glLoadIdentity()
        if self.showFPS:
            glColor3f(1,1,1)
            self.output_text(15,15,'fps: %.2f' % self.fps)
        b = self.highlightBox
        if b:
            glColor4f(b.red,b.green,b.blue,b.alpha)
            glTranslatef(b.x+5,b.y+5,0)
            rectWidth = glutBitmapLength(GLUT_BITMAP_TIMES_ROMAN_10,b.text)
            self.drawRoundedRect(rectWidth + 20,20)
            glDisable(GL_BLEND)
            glColor3f(0,0,0)
            self.output_text(b.x+15,b.y+10,b.text)
            glColor3f(0.1,1,0.1)
            self.output_text(b.x+14,b.y+11,b.text)
            glEnable(GL_BLEND)
        glPopMatrix()
        glMatrixMode(GL_PROJECTION)
        glPopMatrix()
        glMatrixMode(GL_MODELVIEW)
        glEnable(GL_DEPTH_TEST)
        glEnable(GL_LIGHTING)

        
    # The main drawing function. 
    def DrawGLScene(self):
        cl = time.clock()
        try:
            # Clear The Screen And The Depth Buffer
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
            if not self.area:
                return
            if not self.preprocessed:
                self.preprocess()
                self.recomputeCamera()
                self.preprocessed = True
            self.setupCamera()

            glLightfv(GL_LIGHT0,GL_AMBIENT,[0.0,0.0,0.0,1.0])
#            glLightfv(GL_LIGHT0,GL_EMISSION,[0.0,0.0,0.0,1.0])
            glLightfv(GL_LIGHT0,GL_DIFFUSE,[1.0,1.0,1.0,1.0])
            glLightfv(GL_LIGHT0,GL_SPECULAR,[1.0,1.0,1.0,1.0])
            glLightfv(GL_LIGHT0,GL_POSITION,[self.viewX,
                                             self.viewY,
                                             self.viewZ,
                                             1.0])

            w = self.area.getWidth()*10.0
            h = self.area.getHeight()*10.0

            name = 0
            cull = True
#            tilesVisible = Numeric.ones((self.area.getWidth(),
#                                         self.area.getHeight()))
            for i,t in enumerate(self.tiles):
                if not t.getModel():
                    continue
                x = i % self.area.getWidth()
                y = i / self.area.getWidth()
                tx = x*10.0+5.0
                ty = y*10.0+5.0
#                print t.modelName,'at',x,y,t.getModel().boundingBox[1][2]
                glPushMatrix()
                glTranslatef(tx,ty,0.0)
                glRotatef(t.getBearing(),0,0,1)
                if not cull or self.isVisible(t.getModel()):
                    self.handleNode(t.getModel().getRootNode())
                    #print 'not culling',x,y
                else:
                    #print 'culling',t.getModel().getRootNode().name,x,y
#                    tilesVisible[x,y] = False
                    pass
                #self.renderHighlightBoxOutline(t.getModel().boundingBox)
                glPopMatrix()
            #print '-------------\n'

            glPushName(name)
            for d in self.doors:
                if not d.getModel():
                    continue
                glLoadName(name)
                glPushMatrix()
                glTranslatef(d.getX(),d.getY(),0.01)
                if self.highlight == name:
                    (self.highlightBox.x,
                     self.highlightBox.y,
                     self.highlightBox.z) = self.project(0,0,0)
                glRotatef(d.getBearing() * 180/math.pi,0,0,1)
                if not cull or self.isVisible(d.getModel()):
                    self.handleNode(d.getModel().getRootNode(),
                                    selected=(name in self.selected))
                else:
                    #print 'culling',d.getName()
                    pass
                glPopMatrix()
                name += 1

            glColor3f(0.8,0.5,0.5)
            for p in self.placeables:
                if not p.getModel():
                    continue                
                glLoadName(name)
                glPushMatrix()
                glTranslatef(p.getX(),p.getY(),0.01)
                if self.highlight == name:
                    self.highlightBox.x,self.highlightBox.y,self.highlightBox.z = self.project(0,0,0)
                glRotatef(p.getBearing() * 180/math.pi,0,0,1)
                if not cull or self.isVisible(p.getModel()):
                    self.handleNode(p.getModel().getRootNode(),selected=(name in self.selected))
                else:
                    #print 'culling',p.getName()
                    pass
                glPopMatrix()
                name += 1

            glColor3f(0.8,0.4,0.8)
            for p in self.creatures:
                if not p.getModel():
                    continue                
                glLoadName(name)
                glPushMatrix()
                glTranslatef(p.getX(),p.getY(),0.01)
                if self.highlight == name:
                    self.highlightBox.x,self.highlightBox.y,self.highlightBox.z = self.project(0,0,0)
                glRotatef(p.getBearing() * 180/math.pi,0,0,1)
                if p.getModel() and\
                       (not cull or self.isVisible(p.getModel())):
                    self.handleNode(p.getModel().getRootNode(),selected=(name in self.selected))
                else:
                    pass
                glPopMatrix()
                name += 1

            for p in self.players.values():
                self.lock.acquire()
                x = p.x
                y = p.y
                a = (float(p.angle)/225.0)*360.0
                self.lock.release()
                glPushMatrix()
                glTranslatef(x,y,0.0)
                glRotate(a,0,0,1)
                glColor3f(1.0,0.4,0.4)
                glBegin(GL_LINE_LOOP)
                glVertex3f(-0.05,-0.05,0.0)
                glVertex3f(0.0,0.08,0.0)
                glVertex3f(0.05,-0.05,0.0)
                glEnd()
                glPopMatrix()
            glPopName()

            self.drawOverlays()
            self.SwapBuffers()
            
        except KeyboardInterrupt:
            print 'shutting down'
            sys.exit()
        if self.showFPS:
            d = time.clock()-cl
            if d:
                self.fps = 1.0/d

    def InitGL(self,w,h):
        GLWindow.InitGL(self,w,h)
        
    def get_standalone(cls, mod):
        win = None
        class MyApp(wxApp):
            def OnInit(self):
                resourceManager = ResourceManager()
                try:
                    resourceManager.scanGameDir('/Applications/Neverwinter Nights')
                except:
                    resourceManager.scanGameDir('/home/pgorniak/nwn')
                globals.resourceManager = resourceManager
                m = Module(mod)
                area = m.getArea(m.getProperty('Mod_Entry_Area')[1])
                frame = wxFrame(None, -1, "Map", wxDefaultPosition, wxSize(400,400))
                win = MapWindow(frame)
                win.setArea(area)
                frame.Show(True)
                self.SetTopWindow(frame)
                return True
        cls.app = MyApp(0)
        return win
    get_standalone = classmethod(get_standalone)

    def start_standalone(cls):
        cls.app.MainLoop()
    start_standalone = classmethod(start_standalone)
    
    
# Print message to console, and kick off the main to get it rolling.
if __name__ == "__main__":
    if len(sys.argv) < 2:
        print 'usage: ' + sys.argv[0] + ' <modfile>'
        sys.exit(1)
    MapWindow.get_standalone(sys.argv[1])
    MapWindow.start_standalone()
    	
