import globals

class NeverData:
    '''This class is the superclass to a hierarchy of game classes
    that all have properties stored in one or more gff files. Each
    subclass only has to add property list dictionaries and gff
    structs to their superclass, which handles access to the
    properties by name as well as iterating through them.'''

    globalID = 0
    
    def __init__(self):
        if not hasattr(self,'propListDict'): #this is so we can have multiple inheritance
            self.propListDict = {}
            self.gffstructDict = {}
            self.nevereditID = NeverData.globalID
            NeverData.globalID += 1

    def __getitem__ (self, key):
        try:
            sp = self.getProperty (key)
            return sp[1]
#            if sp[2].split(',')[0] == 'CExoLocString':
#                return sp[1].getString ()
#            else:
#                return sp[1]
        except:
            return None
        
    def __setitem__ (self, key, value):
       return self.setProperty (key, value)

#    def __iter__
#    def __delitem__
#    def __len__
#    def __contains__

    def getNevereditId(self):
        return self.nevereditID
    
    def addPropList(self,name,plist,gffstruct):
        '''Add a property list dictionary and a gff struct under
        the given name.'''
        self.propListDict[name] = plist
        self.gffstructDict[name] = gffstruct
        
    def getTargetStructAndLabel(self,root,spec):
        '''traverse a GFFFile according to a specification with
        periods separating structure names and return the target
        structure of the specification as well as the target
        entry label.'''
        structNames = spec.split('.')
        targetStruct = root
        for s in structNames[:-1]:
            targetStruct = targetStruct.getGFFEntry(s)[0]
        return (targetStruct,structNames[-1])
        
    def getProperty(self,label):
        for name,plist in self.propListDict.iteritems():
            if label in plist:
                return [label,
                        self.gffstructDict[name].getInterpretedEntry(label),
                        plist[label]]
        else:
            #try the first one
            return [label,
                    self.gffstructDict.values()[0].getInterpretedEntry(label),
                    self.propListDict.values()[0].get(label,label)]

    def setProperty(self,label,value):
        for name,plist in self.propListDict.iteritems():
            if label in plist:
                (s,t) = self.getTargetStructAndLabel(self.gffstructDict[name],
                                                     label)
                s.setInterpretedEntry(t,value)
        else:
            #try the first one
            (s,t) = self.getTargetStructAndLabel(self.gffstructDict.values()[0],
                                                 label)
            s.setInterpretedEntry(t,value)

    def iterateProperties(self):
        for name,plist in self.propListDict.iteritems():
            plkeys = plist.keys()
            plkeys.sort()
            for label in plkeys:
                (s,t) = self.getTargetStructAndLabel(self.gffstructDict[name],
                                                     label)
                yield [label,s.getInterpretedEntry(t),plist[label]]

    def getPropertyValue(self,label):
        return self.getProperty(label)[1]
    
    def getPortrait(self,size):
        portraitIndex = self.getPropertyValue('PortraitId')
        return globals.resourceManager.getPortraitByIndex(portraitIndex,size)

