#standard lib
import os,os.path,sys
import cStringIO
from sets import Set

#external lib
import Image
import TgaImagePlugin
import JpegImagePlugin
Image._initialized = 1

#neveredit
from Progressor import Progressor
import MDLFile
import KeyFile
import ERFFile
import GFFFile
import Script
import TalkTableFile
import TwoDAFile
import TileSetFile
import Module

RESOURCETYPES = {
    'INVALID' : 0xFFFF,
    'BMP'     : 1,
    'TGA'     : 3,
    'WAV'     : 4,
    'PLT'     : 6,
    'INI'     : 7,
    'TXT'     : 10,
    'MDL'     : 2002,
    'NSS'     : 2009,
    'NCS'     : 2010,
    'ARE'     : 2012,
    'SET'     : 2013,
    'IFO'     : 2014,
    'BIC'     : 2015,
    'WOK'     : 2016,
    '2DA'     : 2017,
    'TXI'     : 2022,
    'GIT'     : 2023,
    'UTI'     : 2025,
    'UTC'     : 2027,
    'DLG'     : 2029,
    'ITP'     : 2030,
    'UTT'     : 2032,
    'DDS'     : 2033,
    'UTS'     : 2035,
    'LTR'     : 2036,
    'GFF'     : 2037,
    'FAC'     : 2038,
    'UTE'     : 2040,
    'UTD'     : 2042,
    'UTP'     : 2044,
    'DFT'     : 2045,
    'GIC'     : 2046,
    'GUI'     : 2047,
    'UTM'     : 2051,
    'DWK'     : 2052,
    'PWK'     : 2053,
    'JRL'     : 2056,
    'UTW'     : 2058,
    'SSF'     : 2060,
    'NDB'     : 2064,
    'PTM'     : 2065,
    'PTT'     : 2066}

RESOURCEEXTENSIONS = dict(zip(RESOURCETYPES.values(),RESOURCETYPES.keys()))

GFFFILETYPES = ['ARE','IFO','BIC','GIT','UTI','UTC','DLG','ITP','UTT',\
                'UTS','GFF','FAC','UTE','UTD','UTP','GIC','GUI','UTM',\
                'JRL','UTW','PTM','PTT']

cache = {}

def resTypeFromExtension(ext):
    return RESOURCETYPES[ext.upper()]

def extensionFromResType(type):
    try:
        return RESOURCEEXTENSIONS[type]
    except KeyError:
        print >>sys.stderr,'ResourceManager: type',type,'unknown resource type'
        return 'UNK'

def nameFromKey(key):
    return key[0].strip('\0') + '.' + extensionFromResType(key[1])

def keyFromName(name):
    parts = name.split('.')
    k1 = parts[0] + (16-len(parts[0]))*'\0'    
    try:
        k2 = resTypeFromExtension(parts[1])
    except KeyError:
        print >>sys.stderr,'ResourceManager: unknown extension "' + parts[1].upper() + '"'
        return (k1,0xFFFF)
    return (k1,k2)

def getResourceFromCache(key):
    return cache.get(key,None)

def addResourceToCache(key,r):
    cache[key] = r

def clearCache():
    cache = {}
    
def clearResourceCacheByExtension(ext):
    t = resTypeFromExtension(ext)
    for key in cache:
        if t == key[1]:
            del cache[key]
            
def interpretResourceContents(key,data):
    r = getResourceFromCache(key)
    if r:
        return r
    #could alternatively go by 'type' header field for many types...
    if extensionFromResType(key[1]) in GFFFILETYPES:
        gff = GFFFile.GFFFile()
        gff.fromFile(cStringIO.StringIO(data),0)
        addResourceToCache(key,gff)
        return gff
    elif extensionFromResType(key[1]) == 'NSS':
        return Script.Script(nameFromKey(key),data)
    elif extensionFromResType(key[1]) == '2DA':
        f = TwoDAFile.TwoDAFile()
        f.fromFile(cStringIO.StringIO(data))
        addResourceToCache(key,f)
        return f
    elif extensionFromResType(key[1]) == 'TGA':
        im = Image.open(cStringIO.StringIO(data))
        addResourceToCache(key,im)
        return im
    elif extensionFromResType(key[1]) == 'MDL':
        f = MDLFile.MDLFile()
        f.fromFile(cStringIO.StringIO(data))
        m = f.getModel()
        addResourceToCache(key,m)
        return m
    elif extensionFromResType(key[1]) == 'SET':
        f = TileSetFile.TileSetFile()
        f.fromFile(cStringIO.StringIO(data))
        addResourceToCache(key,f)
        return f
    else:
        return data

def keyToString(key):
    return key[0].strip('\0') + '.' + extensionFromResType(key[1])

class ResourceManager(Progressor):
    appDir = ''
    def __init__(self):
        Progressor.__init__(self)
        self.clear()

    def clear(self):
        self.keyResourceKeys = {}
        self.dirResourceKeys = {}
        self.hakResourceKeys = {}
        self.resourcesByType = {}
        self.resourcesByName = {}
        self.hakFileNames = []
        self.mainDialogFile = None
        self.modMap = {}
        
    def scanGameDir(self,dir):
        print >>sys.stderr,'initializing resource manager from "',dir,'"',
        sys.stderr.flush()
        self.clear()
        self.setAppDir(dir)
        if not os.path.exists(self.getAppDir()):
            print >>sys.stderr
            print >>sys.stderr,'warning: dir given to resource manager does not exist'
            raise IOError
        c = 0.0
        files = os.listdir(self.getAppDir())
        for f in files:
            if f == 'dialog.tlk':
                self.mainDialogFile = TalkTableFile.TalkTableFile()
                self.mainDialogFile.fromFile(os.path.join(self.getAppDir(),f))
            c += 1
            self.setProgress((c/len(files))*100.0)
            if f[-4:] == '.key':
                print >>sys.stderr,'.',
                sys.stderr.flush()
                self.addKeyFile(os.path.join(self.getAppDir(),f))
            if f == 'override':
                for over in os.listdir(os.path.join(self.getAppDir(),f)):
                    pass #should add override files here
            if f == 'hak':
                for hak in os.listdir(os.path.join(self.getAppDir(),f)):
                    if hak[-4:] == '.hak':
                        #print 'adding hak file',hak
                        #                        self.addHAKFile(os.path.join(self.getAppDir(),f,hak))
                        self.hakFileNames.append(hak)
            if f == 'modules':
                for mod in os.listdir(os.path.join(self.getAppDir(),f)):
                    if mod[-4:] == '.mod':
                        #self.addMODFile(os.path.join(self.getAppDir(),f,mod))
                        self.modMap[os.path.basename(f)] = os.path.join(self.getAppDir(),f)
        self.setProgress(0)
        print >>sys.stderr
        if not self.mainDialogFile:
            print >>sys.stderr,\
                  'Warning: this does not look like a valid NWN dir'

    def getDialogString(self,strref):
        if not self.mainDialogFile:
            print >>sys.stderr,'not initialized, cannot return dialog string'
            return None
        else:
            return self.mainDialogFile.getString(strref)
        
    def getHAKFileNames(self):
        return self.hakFileNames

    def loadHAKsForModule(self,mod):
        haks = mod.getHAKNames()
        print >>sys.stderr,haks,self.hakFileNames
        haks.reverse() #reverse as later haks have lower priority
        for hak in haks:
            if not hak:
                continue
            if hak not in self.hakFileNames:
                print >>sys.stderr,'Could not find hak ' + hak + ' amongst installed haks'
            else:
                print >>sys.stderr,'adding hak',os.path.join(self.getAppDir(),'hak',hak)
                self.addHAKFile(os.path.join(self.getAppDir(),'hak',hak))
                
    def addKeyFile(self,keyf):
        keyFile = KeyFile.KeyFile(self.getAppDir())
        keyFile.fromFile(keyf)
        keyList = keyFile.getKeyList()
        for key in keyList:
            self.keyResourceKeys[key] = keyFile
            if key[1] in self.resourcesByType:
                self.resourcesByType[key[1]].add(key)
            else:
                self.resourcesByType[key[1]] = Set([key])

    def addHAKFile(self,hak):
        hakFile = ERFFile.ERFFile()
        hakFile.fromFile(hak)
        keyList = hakFile.getKeyList()
        for key in keyList:
            self.hakResourceKeys[key] = hakFile
            if key[1] in self.resourcesByType:
                self.resourcesByType[key[1]].add(key)
            else:
                self.resourcesByType[key[1]] = Set([key])

    def getMODPath(self,m):
        return self.modMap.get(m,m)
    
    def addMODFile(self,erf):
        mod = Module.Module(erf)
        self.addModule(mod)
        
    def addModule(self,mod):
        keyList = mod.getKeyList()
        for key in keyList:
            self.dirResourceKeys[key] = mod.getERFFile()
            if key[1] in self.resourcesByType:
                self.resourcesByType[key[1]].add(key)
            else:
                self.resourcesByType[key[1]] = Set([key])
        self.loadHAKsForModule(mod)

    def cleanMODandHAKKeys(self):
        #note that I'm not clearing out the resourcesByType map right now
        #this means that if someone uses the resource manager to look up
        #resources by type after cleaning out, they might get invalid
        #keys that are not actually currently accessible

        #really, we should probably just speed up initialization and
        #re-initialize instead of cleaning out
        self.dirResources = {}
        self.hakResources = {}
        clearCache()
        
    def getRawResource(self,key):
        if key in self.hakResourceKeys:
            return self.hakResourceKeys[key].getRawResource(key)
        elif key in self.dirResourceKeys:
            return self.dirResourceKeys[key].getRawResource(key)
        elif key in self.keyResourceKeys:
            return self.keyResourceKeys[key].getRawResource(key)
        else:
            print >>sys.stderr,'error: cannot find resource for ',key,'in any added lists'
            return None

    def getRawResourceByName(self,name):
        key = keyFromName(name)
        return self.getRawResource(key)

    def getResource(self,key):
        r = getResourceFromCache(key)
        if r:
            return r
        if key in self.hakResourceKeys:
            return self.hakResourceKeys[key].getResource(key)
        elif key in self.dirResourceKeys:
            return self.dirResourceKeys[key].getResource(key)
        elif key in self.keyResourceKeys:
            return self.keyResourceKeys[key].getResource(key)
        else:
            print >>sys.stderr,'error: cannot find resource for ',key,'in any added lists'
            return None

    def getResourceByName(self,name):
        key = keyFromName(name)
        return self.getResource(key)
    
    def getKeysWithExtensions(self,ext):
        return self.resourcesByType[resTypeFromExtension(ext)]

    def getTextScriptKeys(self):
        return self.getKeysWithExtensions('nss')

    def getAppDir(cls):
        return cls.appDir
    getAppDir = classmethod(getAppDir)

    def setAppDir(cls,dir):
        cls.appDir = dir
    setAppDir = classmethod(setAppDir)

    def getPortraitByIndex(self,index,size):
        twoda = self.getResourceByName('portraits.2da')
        name = 'po_' + twoda.getEntry(index,'BaseResRef').lower()\
               + size + '.tga'
        return self.getResourceByName(name)

    def getPortraitNameList(self):
        portraits = Set()
        for key in self.getKeysWithExtensions('TGA'):
            if key[0][:3] == 'po_':
                portraits.add(key[0].strip('\0')[:-1])
        return list(portraits)
    
